
/*
 * Quick and dirty MIDI note mapper
 */

/*
 * Author:  P.J. Drongowski
 * Site:    http://sandsoftwaresound.net/
 * Version: 0.1
 * Date:    11 December 2017
 *
 * Copyright (c) 2017 Paul J. Drongowski
 * Permission granted to modify and distribute
 */

/*
 * This source file contains the main program for the note mapper. It parses
 * command line arguments and invokes the note mapper class.
 *
 * The note mapper is launched from the command line (no GUI). There are two
 * command line options:
 *    -v  Verbose flag
 *    -w  Write default keymap.txt, velmap.txt and hhmap.txt files
 * The -w option gives the user a quick start by writing a few default map
 * files. Both -v and -w are optional.
 *
 * The rest of the command line consists of an (optional) MIDI channel
 * number and the name of the MIDI file to be mapped. The channel number must
 * be an integer in the range [1:16]. Only notes in the specified channel
 * are mapped. The default channel is 10 (the GM/XG drum channel).
 *
 * The note mapped writes a new file named mapped.mid. It's up to the user to
 * rename or save this file. If the file in not renamed or saved, it will be
 * overwritten when note mapper is run again.
 *
 * The note mapper assumes there are two files, keymap.txt and velmap.txt, in
 * the working directory where the note mapper is launched. The note mapper
 * reports an error if it cannot read these two files.
 * 
 * A map file consists of 128 positive integers in the range [0:127].
 * Each integer defines how its corresponding note or velocity value is
 * mapped to a new value. Essentially, each integer in the file is loaded
 * into a 128 byte map array indexed by either the incoming MIDI note
 * number or the incoming MIDI velocity value.
 *
 * The hhmap.txt maps the Revo hi-hat note numbers to General MIDI hi-hat
 * note numbers.
 *     Revo notes        GM notes
 *     -------------     ----------------
 *     13 14 15 16   --> 42 Hi-Hat Closed
 *     21            --> 44 Hi-Hat Pedal
 *     17 18 19 20   --> 46 Hi-Hat Open
 *     22            --> 55 Splash Cymbal
 */

import java.io.File;

import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiFileFormat;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;

public class MapNotes {

  public static void main(String[] args) throws Exception {
    String midiFileName = "" ;
    String outputFileName = "" ;
    boolean verbose = false ;
    boolean writeMapFiles = false ;
    int channelNumber = 10 ;

    if (args.length <= 0) {
      System.err.println("Missing command line arguments") ;
    }

    for (int i = 0 ; i < args.length ; i++) {
      switch( args[i].charAt(0)) {
      case '0' :
      case '1' :
      case '2' :
      case '3' :
      case '4' :
      case '5' :
      case '6' :
      case '7' :
      case '8' :
      case '9' :
          // Channel number
	  channelNumber = Integer.parseInt(args[i]) ;
	  if (verbose) System.out.println("Set channel: " + channelNumber) ;
	  break ;
      case '-' :
	  switch( args[i].charAt(1)) {
	  case 'v' :
	      // Verbose flag
	      verbose = true ;
	      break ;
	  case 'w' :
	      // Write default map files
	      writeMapFiles = true ;
	      break ;
	  default :
	      System.err.println("Unknown option") ;
	      break ;
	  }
	  break ;
      default :
	// Assume it's the MIDI file name
        midiFileName = args[i] ;
	break ;
      }
    }

    outputFileName = "mapped.mid" ;
    // outputFileName = "_" + midiFileName ;

    if (verbose) {
      System.out.println("MIDI file: " + midiFileName) ;
      System.out.println("Output file: " + outputFileName) ;
      System.out.println("Channel: " + channelNumber) ;
      for (String s: args) {
        System.out.println(s);
      }
    }

    if ((channelNumber <= 0) || (channelNumber > 16)) {
      System.err.println("Invalid channel number (1 to 16 required)") ;
      return ;
    }

    if (midiFileName != "") {
      File midiFile = new File(midiFileName) ;
      MidiFileFormat midiFileFormat = MidiSystem.getMidiFileFormat(midiFile) ;
      int fileType = midiFileFormat.getType() ;
      float divisionType = midiFileFormat.getDivisionType() ;
      int resolution = midiFileFormat.getResolution() ;
      int byteLength = midiFileFormat.getByteLength() ;
      long microsecondLength = midiFileFormat.getMicrosecondLength() ;

      if (verbose) {
        System.out.println("File type:     " + fileType) ;
        System.out.println("File length:   " + byteLength) ;
      }

      Sequence sequence = MidiSystem.getSequence(midiFile);
      NoteMapper nm = new NoteMapper(sequence, channelNumber) ;

      if (writeMapFiles) {
	  nm.writeMapFiles() ;
      }

      nm.readMapFiles("keymap.txt", "velmap.txt") ;
      nm.mapNotes() ;
      MidiSystem.write(sequence, fileType, new File(outputFileName)) ;
    }
  }
}
