
/*
 * Simple MIDI note mapper (key number and velocity)
 */

/*
 * Author:  P.J. Drongowski
 * Site:    http://sandsoftwaresound.net/
 * Version: 0.1
 * Date:    11 December 2017
 *
 * Copyright (c) 2017 Paul J. Drongowski
 * Permission granted to modify and distribute
 */

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import java.util.Scanner;

import javax.sound.midi.MidiEvent;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Track;
import javax.sound.midi.InvalidMidiDataException;

public class NoteMapper {
  private Sequence sequence ;
  private int mapChannel = 10 ;
  private byte[] keyMap ;
  private byte[] velocityMap ;

  // No mapping
  // Revo: JazzBrushKitComp
  public static final int NO_MAPPING = 0 ;
  // Revo: RockDrumKit, PopDrumKit, VintageOpenKit, JazzBrushExpanded,
  // VintageMutedKit, JazzStickKit
  public static final int HIHAT_MAPPING = 1 ;

  private static final byte[] noChange = {
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15,
      16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
      32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
      48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
      64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
      80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
      96, 97, 98, 99, 100,101,102,103,104,105,106,107,108,109,110,111,
      112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127
  } ;

  private static final byte[] hiHatMap = {
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 42, 42, 42,
      42, 46, 46, 46, 46, 44, 55, 23, 24, 25, 26, 27, 28, 29, 30, 31,
      32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
      48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
      64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
      80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
      96, 97, 98, 99, 100,101,102,103,104,105,106,107,108,109,110,111,
      112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127
  } ;

  private static final int NOTE_ON = 0x90 ;
  private static final int NOTE_OFF = 0x80 ;
  private static final int CC = 0xB0 ;
  private static final int PC = 0xC0 ;
  private static final int CAT = 0xD0 ;
  private static final int PAT = 0xD0 ;
  private static final int PB = 0xD0 ;
  private static final int MIDI_CLOCK = 0xF8 ;
  private static final int MIDI_START = 0xFA ;
  private static final int MIDI_CONTINUE = 0xFB ;
  private static final int MIDI_STOP = 0xFC ;
  private static final int MIDI_ACTIVE_SENSE = 0xFE ;
  private static final int MIDI_SYSTEM_RESET = 0xFE ;

  private static final String[] NOTE_NAMES = 
  {
    "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"
  } ;

  public NoteMapper(Sequence s, int channel) {
      sequence = s ;
      // Subtract one to convert to internal channel number
      mapChannel = channel - 1 ;
      keyMap = new byte[128] ;
      velocityMap = new byte[128] ;
  }

  public NoteMapper(Sequence s) {
      sequence = s ;
      // Internal representation of MIDI channel 10 is 9
      mapChannel = 9 ;
      keyMap = new byte[128] ;
      velocityMap = new byte[128] ;
  }

  private void handleShortMessage(ShortMessage sm) {
      // Get the channel and return if it's not the channel to be mapped
      int channel = sm.getChannel() ;
      if (channel != mapChannel) return ;

      if (sm.getCommand() == NOTE_ON) {
	  int key = sm.getData1() ;
	  int velocity = sm.getData2() ;
	  int newkey = keyMap[key] ;
	  int newvel = velocityMap[velocity] ;
	  try {
	      sm.setMessage(ShortMessage.NOTE_ON, channel, newkey, newvel) ;
	  } catch (InvalidMidiDataException ex) {
	      System.err.println("setMessage NOTE_ON exception") ;
	  }
      } else if (sm.getCommand() == NOTE_OFF) {
	  int key = sm.getData1() ;
	  int velocity = sm.getData2() ;
	  int newkey = keyMap[key] ;
	  try {
	      sm.setMessage(ShortMessage.NOTE_OFF, channel, newkey, velocity) ;
	  } catch (InvalidMidiDataException ex) {
	      System.err.println("setMessage NOTE_OFF exception") ;
	  }
      } else {
	  // STUB: Other short MIDI message: CC, PC, etc.
      }
  }

  private void scanOneTrack(Track track) {
    for (int i=0; i < track.size(); i++) { 
      MidiEvent event = track.get(i) ;
      MidiMessage message = event.getMessage() ;
      if (message instanceof ShortMessage) {
        ShortMessage sm = (ShortMessage) message ;
	handleShortMessage(sm) ;
      } else {
        // STUB: Handle a long message (SysEx or Meta)
      }
    }
  }

    private void scanTracks() {
	int trackNumber = 0 ;
	for (Track track :  sequence.getTracks()) {
	    trackNumber++;
	    // System.out.println("Track "+trackNumber+" Size:"+track.size()) ;
	    // System.out.println() ;
	    scanOneTrack(track) ;
	}
    }

    public void readMapFiles(String keyFileName, String velFileName) {
	// Read the key (MIDI note number) map
	try {
	    Scanner scanner = new Scanner(new File(keyFileName)) ;
	    int i = 0 ;
	    while(scanner.hasNextInt()) {
		if (i < 128) keyMap[i++] = (byte)scanner.nextInt() ;
	    }
	} catch (IOException ex) {
	    System.err.println("Error reading key map file: " + keyFileName) ;
	    ex.printStackTrace() ;
	}

	// Read the velocity map
	try {
	    Scanner scanner = new Scanner(new File(velFileName)) ;
	    int i = 0 ;
	    while(scanner.hasNextInt()) {
		if (i < 128) velocityMap[i++] = (byte)scanner.nextInt() ;
	    }
	} catch (IOException ex) {
	    System.err.println("Error reading velocity map file: " + velFileName) ;
	    ex.printStackTrace() ;
	}
    }

    public void writeMapFiles() {
	FileWriter fileWriter = null ;

	// Write the default keymap.txt file (no note number changes)
	try {
	    fileWriter = new FileWriter("keymap.txt") ;
	    for (int i = 0; i < noChange.length; i++) {
		fileWriter.write(Integer.toString(noChange[i])) ;
		fileWriter.write(System.lineSeparator()) ;
	    }
	    fileWriter.close() ;
	} catch (IOException ex) {
	    System.err.println("Error writing key map file: keymap.txt") ;
	    ex.printStackTrace() ;
	}

	// Write the default velmap.txt file (no velocity changes)
	try {
	    fileWriter = new FileWriter("velmap.txt") ;
	    for (int i = 0; i < noChange.length; i++) {
		fileWriter.write(Integer.toString(noChange[i])) ;
		fileWriter.write(System.lineSeparator()) ;
	    }
	    fileWriter.close() ;
	} catch (IOException ex) {
	    System.err.println("Error writing velocity map file: velmap.txt") ;
	    ex.printStackTrace() ;
	}

	// Write the default hhmap.txt file (Revo Hi-Hat changes)
	try {
	    fileWriter = new FileWriter("hhmap.txt") ;
	    for (int i = 0; i < hiHatMap.length; i++) {
		fileWriter.write(Integer.toString(hiHatMap[i])) ;
		fileWriter.write(System.lineSeparator()) ;
	    }
	    fileWriter.close() ;
	} catch (IOException ex) {
	    System.err.println("Error writing hi-hat map file: hhmap.txt") ;
	    ex.printStackTrace() ;
	}
    }

    public void mapNotes() {
	scanTracks() ;
    }
}
